#!/usr/bin/perl -wTC7

# (c)2003 Christopher Hofmann <cwh@suse.de>
# (c)2006 Michael Rumpf <michael@rumpfonline.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#######################################################################
#
# This successor version of the SMPPPD-WEB interface by Christopher
# Hofmann adds the functionality to switch providers and to add and
# remove links for ISDN interfaces. Essentially this HTML interface
# to the SMPPPD now provides the nealry same functionality as the
# KInternet application.
# Internally a code cleanup has been done which might serve as a
# starting point to create a Perl Module for SMPPPD functions.
#
#######################################################################

use strict;
use IO::Socket;
use CGI;
use CGI::Carp 'fatalsToBrowser';
use HTML::Template;
use Log::Log4perl qw(:easy);
use Locale::gettext;
use POSIX;

use vars qw($conn);

# Config
# -----------------------------------------------------------
my $HOST = "127.0.0.1";
my $PORT = 3185;
my $DBG  = 0;
my $REFRESH = 5; #Disable: -1

# Init
# -----------------------------------------------------------
Log::Log4perl->easy_init(
    category => 'main',
    level => $DBG ? $DEBUG : $INFO,
    layout => "%m"
    );

# TODO: Fix getting locale from environment
my $locale = setlocale(LC_MESSAGES, "de_DE.UTF8");
my ($LANG, $CHARSET) = split(/\./, $locale);

my $dom = Locale::gettext->domain("smpppd-web");
$dom->dir("locale");

# CGI Main
# -----------------------------------------------------------
DEBUG "\n\n\n\nSMPPPD-WEB V3.0 ($LANG)\n\n";

&SMPPPDConnect();
&SMPPPDParseCGIRequest();

&SMPPPDCreateHTMLResponse();
&SMPPPDDisconnect();

#######################################################################
#
# smpppd-web uses the following commands from the SMPPPD.
#
# GLOBAL LEVEL ________________________________________________________
#
#   quit
#     quit from the demon
#
#   survive-me 0|1
#     1 : Makes the current connection to survive a disconnect
#         from the SMPPPD
#     0 : Hangup current connection upon quit command
#
#   list-ifcfgs
#     FLAG "INTERFACE" "MENU"
#     FLAG = i|a
#        i : no backend (?)
#        a : backend present (?)
#     INTERFACE = interface name
#     MENU = display name for menus
#
#
# INTERFACE LEVEL _____________________________________________________
#
#   list-status INTERFACE
#     KEY VALUE
#        status [disconnected|lurking|connected|connecting|disconnecting]
#        action-flags 0x0000
#        number-of-links 0..2
#
#   list-providers INTERFACE
#     FLAG PROVIDER DESCRIPTION
#     FLAG = i|a
#        i : inactive
#        a : active
#     PROVIDER = provider identifier
#     DESCRIPTION = free text description of the provider
#
#   list-config INTERFACE
#     connection-type isdn|modem|wlan
#     ppp-mode ippp
#     device "ippp0"
#     username "arcor"
#     password <hidden>
#     ask-password no
#     demand yes
#     max-number-of-links 2
#
#   list-log INTERFACE
#
#   start INTERFACE [PASSWORD]
#
#   stop INTERFACE
#
#   select-provider INTERFACE PROVIDER
#
#   add-link INTERFACE
#
#   remove-link INTERFACE
#
#   dialin INTERFACE
#
#   hangup INTERFACE
#

#######################################################################
# CGI functions
#######################################################################

# Create the HTML response for the CGI request
# -----------------------------------------------------------
sub SMPPPDCreateHTMLResponse
{
    my @ifcfg_ref_list  = &SMPPPDListInterfaces();
    my @ifcfgs;
    foreach my $ifcfg_ref (@ifcfg_ref_list)
    {
        my @ifcfg_arr = @$ifcfg_ref;
        DEBUG "### Interface: " . $ifcfg_arr[1] . " (" . $ifcfg_arr[2] . ")";

        my $status = &SMPPPDItfGetConnectionStatus($ifcfg_arr[1]);
        my $cur_links = &SMPPPDItfGetNumberOfLinks($ifcfg_arr[1]);
        my $max_links = &SMPPPDItfGetMaximumNumberOfLinks($ifcfg_arr[1]);
        my $dod =  &SMPPPDItfIsDialOnDemand($ifcfg_arr[1]);
        my @providers = &SMPPPDItfGetAlternateProviders($ifcfg_arr[1]);
        my $provider = &SMPPPDItfGetCurrentProviderName($ifcfg_arr[1]);
        my $logs = &SMPPPDItfGetLogs($ifcfg_arr[1]);

        my $links = $cur_links." / ".$max_links;
        my @actions;
        my $image_link = '';
        my $status_name = "";
        if ( $status =~ "disconnected" )
        {
            if ($dod)
            {
                # start
                $image_link = "images/disconnected_dod";
                my %start_link = ( action_link => "?action=start&ifcfg=".$ifcfg_arr[1],
                                   action_name => $dom->get("Starten"));
                push(@actions, \%start_link);
            }
            else
            {
                # connect
                $image_link = "images/disconnected";
                my %start_link = ( action_link => "?action=start&ifcfg=".$ifcfg_arr[1],
                                   action_name => $dom->get("Verbinden"));
                push(@actions, \%start_link);
            }
            $status_name = $dom->get("Gestoppt");
        }
        elsif ($status =~ "lurking")
        {
            $image_link = "images/lurking_dod";
            # stop
            my %stop_link = ( action_link => "?action=stop&ifcfg=".$ifcfg_arr[1],
                              action_name => $dom->get("Stoppen"));
            push(@actions, \%stop_link);
            # dialin
            my %dialin_link = ( action_link => "?action=dialin&ifcfg=".$ifcfg_arr[1],
                                action_name => $dom->get("W&auml;hlen"));
            push(@actions, \%dialin_link);
            $status_name = $dom->get("Wartend");
        }
        elsif ($status =~ "connected")
        {
            # stop
            my %stop_link = ( action_link => "?action=stop&ifcfg=".$ifcfg_arr[1],
                              action_name => $dom->get("Stoppen"));
            push(@actions, \%stop_link);

            if ($dod)
            {
                $image_link = "images/connected_dod";

                # hangup
                my %hangup_link = ( action_link => "?action=hangup&ifcfg=".$ifcfg_arr[1],
                                    action_name => $dom->get("Auflegen"));
                push(@actions, \%hangup_link);
            }
            else
            {
                $image_link = "images/connected";
            }

            # Only if maximum number of connections is not reached
            if ($cur_links lt $max_links)
            {
                my %add_link = ( action_link => "?action=add&ifcfg=".$ifcfg_arr[1],
                                 action_name => $dom->get("Kanal hinzuf&uuml;gen"));
                push(@actions, \%add_link);
            }

            # Only if number of links is greater than 1
            if ($cur_links gt 1)
            {
                my %remove_link = ( action_link => "?action=remove&ifcfg=".$ifcfg_arr[1],
                                    action_name => $dom->get("Kanal entfernen"));
                push(@actions, \%remove_link);

                # show an icon indicating that there is more than one link
                $image_link = $image_link."_link";
            }
            $status_name = $dom->get("Verbunden");
        }
        else
        {
            # Unknown status
        }

        $image_link = $image_link.".gif";
        my %interface = (
                          menu_name => $ifcfg_arr[2],
                          name => $ifcfg_arr[1],
                          status => $status_name,
                          image => $image_link,
                          links => $links,
                          disabled => $status !~ "disconnected",
                          provider => $provider,
                          providers => \@providers,
                          actions => \@actions,
                          logs => $logs
                        );
        push(@ifcfgs, \%interface);
    }

    my $tmpl_file = "locale/".$LANG."/smpppd-web.tmpl";
    my $tmpl = HTML::Template->new( filename => $tmpl_file,
                                    cache => 1,
                                    debug => $DBG,
                                    global_vars => 1,
                                    die_on_bad_params => 1 );

    $tmpl->param('ifcfg_loop' => \@ifcfgs);
    $tmpl->param('refresh' => $REFRESH);
    $tmpl->param('script_name' => "smpppd-web.pl");
    $tmpl->param('revision' => '$Revision: 188 $');

    print "Content-type: text/html\n\n";
    print $tmpl->output();

}

# Parse CGI-parameters from request
# -----------------------------------------------------------
sub SMPPPDParseCGIRequest
{
    my $query  = new CGI;

    DEBUG ">>> SMPPPDParseCGIRequest()";

    my $provider = $query->param('select-provider') || '';
    my $ifcfg = $query->param('ifcfg') || '';
    if ($provider)
    {
        DEBUG "### Selecting provider: $ifcfg $provider";
        &SMPPPDSelectProvider($ifcfg, $provider);
    }
    else
    {
        my $action = $query->param('action') || '';
        my $dod = 0;
        if ( $ifcfg ne '' )
        {
            $dod =  &SMPPPDItfIsDialOnDemand($ifcfg);
        }

        if( $action eq 'start' )
        {
            DEBUG "### Starting interface: $ifcfg";
            &SMPPPDItfStart($ifcfg);
        }
        elsif( $action eq 'stop' )
        {
            DEBUG "### Stopping interface: $ifcfg";
            &SMPPPDItfStop($ifcfg);
        }
        elsif( $action eq 'dialin' )
        {
            # Only when in dial on demand mode
            if ( $dod)
            {
                DEBUG "### Dialing interface: $ifcfg";
                &SMPPPDItfDialin($ifcfg);
            }
        }
        elsif( $action eq 'hangup' )
        {
            # Only when in dial on demand mode
            if ( $dod)
            {
                DEBUG "### Hangup interface: $ifcfg";
                &SMPPPDItfHangup($ifcfg);
            }
        }
        elsif( $action eq 'add' )
        {
            # Add an ISDN link (TODO: ERROR - Only one line available)
            DEBUG "### Adding link to interface: $ifcfg";
            &SMPPPDItfAddLink($ifcfg);
        }
        elsif( $action eq 'remove' )
        {
            # Remove an ISDN link (TODO: ERROR - No line available)
            DEBUG "### Removing link to interface: $ifcfg";
            &SMPPPDItfRemoveLink($ifcfg);
        }
        else
        {
            # Unknown action or no action
        }
    }
}

#######################################################################
# Interface-level functions
#######################################################################

# Checks whether the interface has a backend (?)
# -----------------------------------------------------------
sub SMPPPDItfHasBackend
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfHasBackend($itf)";

    my $result = 0;
    my @ref_list = &SMPPPDListInterfaces($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0].", ".$param[1].", ".$param[2];
        if ( $param[0] =~ "a" && $param[1] =~ $itf)
        {
           $result=1;
        }
    }

    DEBUG "<<< SMPPPDItfHasBackend($itf)";

    return $result;
}

# Checks whether DialOnDemand is active for this interface
# -----------------------------------------------------------
sub SMPPPDItfIsDialOnDemand
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfIsDialOnDemand($itf)";
    my @ref_list = &SMPPPDItfListConfig($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "demand" && $param[1] =~ "yes" )
        {
           return 1;
        }
    }
    return 0;
}

# Return the current connection status
# -----------------------------------------------------------
sub SMPPPDItfGetConnectionStatus
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfGetConnectionStatus($itf)";
    my @ref_list = &SMPPPDItfListStatus($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "status" )
        {
           return $param[1];
        }
    }
}

# Return the maximum number of links
# -----------------------------------------------------------
sub SMPPPDItfGetMaximumNumberOfLinks
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfGetMaximumNumberOfLinks($itf)";
    my @ref_list = &SMPPPDItfListConfig($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "max-number-of-links" )
        {
           return $param[1];
        }
    }
}

# Return the current number of links
# -----------------------------------------------------------
sub SMPPPDItfGetNumberOfLinks
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfGetNumberOfLinks($itf)";
    my @ref_list = &SMPPPDItfListStatus($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "number-of-links" )
        {
           return $param[1];
        }
    }
}

# Return the current action flags
# -----------------------------------------------------------
sub SMPPPDItfGetActionFlags
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfGetActionFlags($itf)";
    my @ref_list = &SMPPPDItfListStatus($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "action-flags" )
        {
           return $param[1];
        }
    }
}

# Return the current provider value
# -----------------------------------------------------------
sub SMPPPDItfGetCurrentProviderValue
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfGetCurrentProviderValue($itf)";
    my @ref_list = &SMPPPDItfListProviders($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "a" )
        {
           return $param[1];
        }
    }
}

# Return the current provider name
# -----------------------------------------------------------
sub SMPPPDItfGetCurrentProviderName
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfGetCurrentProviderName($itf)";
    my @ref_list = &SMPPPDItfListProviders($itf);
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "a" )
        {
           return $param[2];
        }
    }
}

# Return the alternate providers
# -----------------------------------------------------------
sub SMPPPDItfGetAlternateProviders
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfListProviders($itf)";
    my @ref_list = &SMPPPDItfListProviders($itf);
    my @providers = ();
    for my $ref ( @ref_list )
    {
        my @param = @$ref;
        DEBUG "### PARAM: ".$param[0]."=".$param[1];
        if ( $param[0] =~ "i" )
        {
           my %provider = ( "provider_value" => $param[1],
                            "provider_name" => $param[2] );
           push(@providers, \%provider);
        }
    }
    return @providers;
}

# Return the log for the interface
# -----------------------------------------------------------
sub SMPPPDItfGetLogs
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfListLog($itf)";
    my @ref_list = &SMPPPDItfListLog($itf);
    my $logs = "";
    for my $ref ( @ref_list )
    {
        $logs = $logs . "\n" . $ref;
    }
    return $logs;
}

#
# Commands
#

# Select a new provider
# -----------------------------------------------------------
sub SMPPPDSelectProvider
{
    my($ifcfg, $provider) = @_;

    DEBUG ">>> SMPPPDSelectProvider($ifcfg, $provider)";
    &SMPPPDSendCommand("select-provider $ifcfg $provider");
    # TODO: Error Handling
}

# Start the interface
# -----------------------------------------------------------
sub SMPPPDItfStart
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfStart($itf)";
    &SMPPPDSendCommand("start $itf");
    # TODO: Error Handling
}

# Dialin the interface
# -----------------------------------------------------------
sub SMPPPDItfDialin
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfDialin($itf)";
    &SMPPPDSendCommand("dialin $itf");
    # TODO: Error Handling
}

# Hangup the interface
# -----------------------------------------------------------
sub SMPPPDItfHangup
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfHangup($itf)";
    &SMPPPDSendCommand("hangup $itf");
    # TODO: Error Handling
}

# Stop the interface
# -----------------------------------------------------------
sub SMPPPDItfStop
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfStop($itf)";
    &SMPPPDSendCommand("stop $itf");
    # TODO: Error Handling
}

# Adding a link to the interface
# -----------------------------------------------------------
sub SMPPPDItfAddLink
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfAddLink($itf)";
    &SMPPPDSendCommand("add-link $itf");
    # TODO: Error Handling
}

# Removing a link to the interface
# -----------------------------------------------------------
sub SMPPPDItfRemoveLink
{
    my($itf) = @_;

    DEBUG ">>> SMPPPDItfRemoveLink($itf)";
    &SMPPPDSendCommand("remove-link $itf");
    # TODO: Error Handling
}

#######################################################################
# Interface-level List functions
#######################################################################

# List the log information for the specified interface
# -----------------------------------------------------------
sub SMPPPDItfListLog
{
    my($itf) = @_;

    &SMPPPDSendCommand("list-log $itf");
    my @ref_list = &SMPPPDReadResponse("LOG");
    return @ref_list;
}

# List the configuration for the specified interface
# -----------------------------------------------------------
sub SMPPPDItfListConfig
{
    my($itf) = @_;

    &SMPPPDSendCommand("list-config $itf");
    my @ref_list = &SMPPPDReadResponse("CONFIG");
    return @ref_list;
}

# List the status of the specified interface
# -----------------------------------------------------------
sub SMPPPDItfListStatus
{
    my($itf) = @_;

    &SMPPPDSendCommand("list-status $itf");
    my @ref_list = &SMPPPDReadResponse("STATUS");
    return @ref_list;
}

# List all configured providers for the specified interface
# -----------------------------------------------------------
sub SMPPPDItfListProviders
{
    my($itf) = @_;

    &SMPPPDSendCommand("list-providers $itf");
    my @ref_list = &SMPPPDReadResponse("PROVIDERS");
    return @ref_list;
}

# List all interfaces
# -----------------------------------------------------------
sub SMPPPDListInterfaces
{
    &SMPPPDSendCommand("list-ifcfgs");
    my @ref_list = &SMPPPDReadResponse("IFCFGS");
    return @ref_list;
}

#######################################################################
# Low-level functions
#######################################################################

# Connects to smpppd
# -----------------------------------------------------------
sub SMPPPDConnect
{
    DEBUG ">>> SMPPPDConnect(".@_.")";

    $/ = "\r\n";

    # Some debug info
    if(exists $ENV{MOD_PERL})
    {
        # WARN "Running under mod_perl\n";
    }
    else
    {
        # WARN "Not running under mod_perl\n";
    }

    # Check for existing connection and connect if necessary
    if( !defined( $conn ) || !$conn->connected() )
    {
        DEBUG "### Connecting to $HOST:$PORT";

        $conn = IO::Socket::INET->new( Proto => "tcp",
                       PeerAddr => $HOST,
                       PeerPort => $PORT )
        or die($dom->get("Verbindung fehlgeschlagen ").$HOST.":".$PORT."\n");

        # Flush TCP buffer after every single line sent
        $conn->autoflush(1);

        # Read the version string
        my $version = <$conn>;
        chomp($version);
        DEBUG "### Server hello: '$version'";

        # Tell smpppd not ho hang up when no client is connected
        &SMPPPDSendCommand("survive-me 1");
    }
    else
    {
        DEBUG "### Reusing existing connection";
    }

    DEBUG "<<< SMPPPDConnect()";
}

# Disconnects from smpppd
# -----------------------------------------------------------
sub SMPPPDDisconnect
{
    DEBUG ">>> SMPPPDDisconnect(".@_.")";

    # Shut down connection when not running under mod_perl
    # because then there is no persistent connection to smpppd
    # -----------------------------------------------------------
    if(exists $ENV{MOD_PERL})
    {
        DEBUG "### Keeping alive connection to $HOST:$PORT";
    }
    else
    {
        &SMPPPDSendCommand( "quit" );
        DEBUG "### Shutting down connection to $HOST:$PORT";
        $conn->shutdown(2);
    }

    DEBUG "<<< SMPPPDDisconnect()";
}

# Send commands to smpppd
# -----------------------------------------------------------
sub SMPPPDSendCommand
{
    DEBUG ">>> SMPPPDSendCommand(".@_.")";

    my($cmd) = @_;

    DEBUG "### Send command: '".$cmd."'";
    print $conn $cmd."\n";

    DEBUG "<<< SMPPPDSendCommand()";
}

# Reads a response from smpppd
# -----------------------------------------------------------
sub SMPPPDReadResponse
{
    DEBUG ">>> SMPPPDReadResponse(".@_.")";

    #$/ = "\r\n";
    my($pattern) = @_;

    my @lines = ();

    # read until /^BEGIN/
    my $line;
    do
    {
        $line = <$conn>;
        chomp ($line);
        DEBUG "### line='".$line."'";
    }
    while ( $line !~ /^BEGIN/ );
    #while ( defined($line) && $line !~ /^BEGIN/ );

    # TODO: Detect "error: ..." message

    if ( $line =~ /^BEGIN\s+(.+)\s+/ )
    {
        my $name = $1;
        if ( $name =~ $pattern )
        {
            do
            {
                $line = <$conn>;
                chomp($line);

                if ( not $line =~ /^END\s+(.+)/ )
                {
                    push(@lines, &splitLine($line));
                }
            }
            while ( not $line =~ /^END\s+(.+)/ );
        }
        else
        {
           die($dom->get("Unerwartete Antwort vom Server '").$name.$dom->get("', erwartet '").$pattern."'!");
        }
    }
    else
    {
        die($dom->get("Unerwartete Antwort vom Server '").$line."'");
    }

    DEBUG "<<< SMPPPDReadResponse()";

    return @lines;
}

#######################################################################
# Helper functions
#######################################################################

sub splitLine
{
    my($line) = @_;

    #DEBUG "Splitting line: >$line<";

    my @options;
    # We either have three parameter lines
    if ( $line =~ /^(.+) (\".*\") (\".*\")$/ )
    {
        DEBUG "### READ: >$1< >$2< >$3<";

        push @options, &stripQuotes($1);
        push @options, &stripQuotes($2);
        push @options, &stripQuotes($3);
    }
    # or two parameter lines
    elsif ( $line =~ /^(.+) (\".*\")$/ or
            $line =~ /^(.+) (.*)$/ )
    {
        DEBUG "### READ: >$1< >$2<";

        push @options, &stripQuotes($1);
        push @options, &stripQuotes($2);
    }

    return \@options;
}

sub stripQuotes
{
  my($option) = @_;

  my $result = $option;
  if ( $option =~ /^\"(.*)\"$/ )
  {
     $result = $1;
  }
  return $result;
}

